// To play SFX, must first import Java sound lib 
// Doing a wildcard import to avoid importing each class one-by-one
import javax.sound.sampled.*;

// To check and open audio files and handle file errors
import java.io.File;
import java.io.IOException;

// For  input
import java.util.Scanner;

public class MagicSquareGame {
	// No need to recreated 
    private static final String PURPLE = "\u001B[35m";
    private static final String RESET = "\u001B[0m";

    public static void printTitle() {
        //String PURPLE = "\u001B[35m"; // ANSI escape code for purple
       // String RESET = "\u001B[0m";   // Reset color
	
		// Store ASCII art as an array
        String[] asciiArt = { // Store ASCII art as an array of lines
            "█ █      █ █         ██         █ █ █ █ █ █ █   █   █ █ █ █ █ █ █      █ █ █ █ █ █   █ █ █ █ █ █ █      █           █         ██         █ █ █ █ █ █    █ █ █ █ █ █ ",
            "█  █    █  █        █  █        █         █ █   █   █           █      █             █           █      █           █        █  █        █          █   █",
            "█   █  █   █       █    █       █               █   █                  █             █           █      █           █       █    █       █          █   █",
            "█    ██    █      █ █  █ █      █               █   █                  █ █ █ █ █ █   █           █      █           █      █ █  █ █      █ █ █ █ █ █    █ █ █ █ █ █ ",
            "█          █     █        █     █       █ █ █   █   █                            █   █           █      █           █     █        █     █          █   █",
            "█          █    █          █    █         █ █   █   █           █                █   █           █      █           █    █          █    █          █   █",
            "█          █   █            █   █ █ █ █ █ █ █   █   █ █ █ █ █ █ █      █ █ █ █ █ █    █ █ █ █ █ █ █ █   █ █ █ █ █ █ █   █            █   █          █   █ █ █ █ █ █"
        };
	
	
		// Set the color of the text to purple
        System.out.println(PURPLE); // Set text color to purple
	
		// Iterate over each line in the ASCII art array print it with a typing effect
        for (String line : asciiArt) { 
            typewriterEffect(line); 
        }
		
		// Reset color after printing
        System.out.println(RESET); 
    }

    // The purpose of this function is to make the game seem more dynamic
    public static void typewriterEffect(String text) {
    	// Setting a fixed delay of 10ms to achieve typing effect 
    	// (Storing in var to make code more readable)
    	int delay = 10;
    	
        for (int i = 0; i < text.length(); i++) {
            System.out.print(text.charAt(i));
            try {
            	// 
                Thread.sleep(delay);
            // Highly unlikely that the current sleeping thread will be interrupted
            // But the Thread class in Java requires error handling
            } catch (InterruptedException e) {
            	// Good practice to invoke interrupt() so the thread doesn't lose the info that was interrupted
                Thread.currentThread().interrupt();
            }
        }
        System.out.println();
    }

    public static void playSound(String soundFile) {
        try {
            File file = new File(soundFile);
            if (!file.exists()) return;  // Fails silently if file is missing
			
			// AudioInputStream represents an audio file that is being read as a stream of bytes 
            AudioInputStream audioStream = AudioSystem.getAudioInputStream(file);
            
            // Creating an audio player for playback
            Clip clip = AudioSystem.getClip();
            
            // Loafing audio into clip
            clip.open(audioStream);
            
            // Playing audio
            clip.start();
			
			// Halt other threads until sound finishes playing
			// Need to divide by 1000 since Thread.sleep takes milliseconds not microseconds (1000 microseconds fit into one millisecond)
            Thread.sleep(clip.getMicrosecondLength() / 1000); 
        
        } catch (Exception ignored) {
        	// Fault masking -- Silently fail
        	// Sound wasn't a requirement to begin with, so the player doesn't need to know that there's smth w/ sound file
        }
    }

    public static void clearScreen() {
    	// Don't need exception handling b/c clearing the screen wasn't a req in the first place
    	// And it's highly unlikely for the ANSI escape seq to fail (except w/ Windows CMD/ some IDEs like Eclipse)
    	System.out.print("\033[H\033[2J");
    	
    	// Flushing the buffer to clear the screen immediately (ensure there are no glitches)
        System.out.flush();
  	}

    public static void main(String[] args) {
    	// Reading  input from terminal
        Scanner scanner = new Scanner(System.in);
        
        // Flag to control game loop
        boolean playAgain = true;
		
        while (playAgain) {
            clearScreen();
            printTitle();
            typewriterEffect("Welcome to the Magic Square Game!");
            typewriterEffect("Your goal is to reconstruct the magic square.");

            System.out.println("\n========== PART A: MAGIC SQUARE GENERATION ==========");
            int size;
            while (true) {
                System.out.print("Enter an odd integer for the magic square size: ");
                
                // If the 's input is not an integer, display an error message and consume the invalid input
                if (!scanner.hasNextInt()) {
                    System.out.println("Invalid input. Please enter a positive odd integer.");
                    scanner.next();
                    
                    // Skip the rest of the loop, go back to the start of the loop
                    continue; 
                }
				// Read integer input and check whether it's both pos&&odd, in which case you can exit the loop
                size = scanner.nextInt();
                if (size % 2 == 1 && size > 0) break;
                System.out.println("Invalid input. Please enter a positive odd integer.");
            }
			
			// Instantiate a new MagicSquare object and display the magic square 
			// This is what lets us use the logic in MagicSquare.java
            MagicSquare magicSquare = new MagicSquare(size);
            System.out.println("\nGenerated Magic Square:");
            magicSquare.display();

		
            System.out.print("\nPress any key to continue to Part B...");
            
            //  First one consumes the leftover newline from  input
            scanner.nextLine();
            
            // Second one waits for  input before moving to Part B
            scanner.nextLine();
            clearScreen();

            System.out.println("\n========== PART B: MAGIC SQUARE GAME ==========\n");
            magicSquare.display();

            System.out.println("\nShuffling the Magic Square...\n");
            try { Thread.sleep(300); } catch (InterruptedException e) { Thread.currentThread().interrupt(); }
            playSound("shuffle.wav");
            magicSquare.shuffleSquare();
            
            System.out.println("\nShuffled Magic Square:\n");
			magicSquare.display(); 
			System.out.println("\nTo win, simply reconstruct the original magic square!");

			
            int moves = 0;
            int moveLimit = 20;
			
			// Loop runs until player wins/ runs out of moves
            while (!magicSquare.isMagicSquare() && moves < moveLimit) {
                System.out.println("\nMoves Left: " + (moveLimit - moves));
                System.out.print("Enter move (row col direction [U/D/L/R] or X to exit):");				
                
				// Initializing variables as unset
                boolean validInput = false;
                int row = -1, col = -1, direction = -1;
                
                while (!validInput) {
                	
                	// Read the entire line and trim extra spaces
        			String inputLine = scanner.nextLine().trim(); // Read full line, remove extra spaces

        			// Deal w/ player wanting to exit first
        			if (inputLine.equalsIgnoreCase("X")) {
            			System.out.println("\nThanks for playing! Goodbye!");
            			return;
        			}

        			// Split input into parts (using whitespace regex)
        			String[] parts = inputLine.split("\\s+"); 

        			// The user's input must consist of 3 parts (row, col, dir), otherwise invalid
        			if (parts.length != 3) {
            			System.out.println("Invalid input. Please enter row, column, and direction (or X to exit).");
            			continue;
        			}

        			// To validate row input, first check if it's even a number then ensure it's non-negative
        			if (!parts[0].matches("-?\\d+")) {
            			System.out.println("Invalid row. Must be a number.");
            			continue;
        			}
        			
        			row = Integer.parseInt(parts[0]) - 1;
        			if (row < 0 || row >= size) {
            			System.out.println("Invalid row. Must be between 1 and " + size);
            			continue;
        			}

        			// Use same logic for column input validation
        			if (!parts[1].matches("-?\\d+")) {
            			System.out.println("Invalid column. Must be a number.");
            			continue;
        			}
        			
        			col = Integer.parseInt(parts[1]) - 1;
        			if (col < 0 || col >= size) {
            			System.out.println("Invalid column. Must be between 1 and " + size);
            			continue;
        			}
        			
                	// 🔹 Check if Direction is valid (U, D, L, R)
        			char dir; 
        			if (parts[2].length() == 1) {
        			// use charAt(0) since toUpperCase() returns a string
        				dir = parts[2].toUpperCase().charAt(0);
        			} else {
        				System.out.println("Invalid direction. Use U, D, L, or R.");
        				continue;
        			}
        			
        			switch (dir) {
    					case 'U': direction = 0; break;
    					case 'D': direction = 1; break;
    					case 'L': direction = 2; break;
    					case 'R': direction = 3; break;
    					default: direction = -1;
					}

        			if (direction == -1) {
            			System.out.println("Invalid direction. Use U, D, L, or R.");
            			continue;
        			}
					
					// If we reach this point, then the input must be valid
        			validInput = true; 
    			}
				
				// Important that we call swapElement with false (swap does NOT happen during shuffling)
    			if (magicSquare.swapElement(row, col, direction, false)) {
        			moves++;
        			System.out.println();
        			magicSquare.display();
    			}
			}
			
			// Check whether the player 
            if (magicSquare.isMagicSquare()) {
            	
            	// Reporting the number of moves as required
                System.out.println("\nYou solved the magic square in " + moves + " moves!");
               
                playSound("victory.wav");
            } else {
                System.out.println("\nGame Over! You ran out of moves.");
                playSound("defeat.wav");
            }

			// Giving player option to replay and validating their response
			String response;
			while (true) {
				System.out.print("\nWould you like to play again? (Y/N): ");
				response = scanner.next().trim().toUpperCase();
				
				if (response.equalsIgnoreCase("Y")){
					playAgain = true;
					break;
				} else if (response.equalsIgnoreCase("N")){
					playAgain = false;
					break;
				} else{
					System.out.println("Invalid input. Please enter either 'Y' for yes or 'N' for no.");
				}
			}
        }
		
        System.out.println("\nThanks for playing! Goodbye!");
        scanner.close();
    }
}